import com.intellij.database.model.DasTable
import com.intellij.database.util.Case
import com.intellij.database.util.DasUtil
import groovy.json.JsonSlurper

import java.text.SimpleDateFormat
import java.util.stream.Collectors


/**
 * 这里需要配置，主要配置项目的根目录，jDaoMySQL会将生成的代码放在固定的包下
 *
 * 例如您的项目的根目录位置在D:\project\java\zclServer，则需要配置的
 * baseDir="D:\\project\\java\\zclServer"
 */
//baseDir = "D:\\project\\java\\monitor-server-uni\\common";
baseDir = "D:\\project\\java\\ops-server";
/**
 * 日志路径，默认日志路径为：D:\jDaoMySQL.log
 */
logPath = "D:\\jDaoMySQL.log";

log("start jCoder++++++++++++++++")

if (!checkVersion(1.0)) {
    throw new Exception("groovy script expired, please download a newer version");
    return
}

log("check version over++++++++++++++++")

typeMapping = [
        (~/(?i)tinyint|smallint|int/): "Integer",
        (~/(?i)bigint/)              : "Long",
        (~/(?i)float|double/)        : "Double",
        (~/(?i)decimal/)             : "BigDecimal",
        (~/(?i)datetime|timestamp/)  : "LocalDateTime",
        (~/(?i)date/)                : "LocalDate",
        (~/(?i)time/)                : "LocalTime",
        (~/(?i)/)                    : "String"
]


if (!baseDir.endsWith(File.separator))
    baseDir = baseDir + File.separator

//读取项目的基础配置
def jsonSlpuer = new JsonSlurper()
def conf = jsonSlpuer.parseText(new File(new File(baseDir), "jcoder.json").text)
sdkModuleFolder = baseDir + conf.sdkModuleFolder
sdkModulePackage = conf.sdkModulePackage

exceptions = new ArrayList<String>();

SELECTION.filter { it instanceof DasTable }.each {
    def dir = sdkModuleFolder +
            File.separator + "src" +
            File.separator + "main" +
            File.separator + "java";
    log("dir:" + dir);
    def entityClassFullName = generateEntity(it, dir)
    def repositoryClassFullName = generateRepository(it, dir, entityClassFullName)
    def managerClassFullName = generateManager(it, dir, entityClassFullName, repositoryClassFullName)
    def voClassFullName = generateVo(it, dir, "Vo");
    def baseVoClassFullName = generateVo(it, dir, "BaseVo");
    def listVoClassFullName = generateVo(it, dir, "ListVo");
    if (exceptions.size() > 0) {
        throw new Exception(exceptions.stream().collect(Collectors.joining(", ")));
    }

}

def log(message) {
    def logName = logPath;
    FileWriter writer = new FileWriter(logName, true);
    writer.write(message);
    writer.write("\r\n")
    writer.close()
}

/**
 * 生成实体
 * @param table
 * @param dir
 * @return
 */
def generateEntity(table, dir) {
    def className = camelCase(table.getName(), true) + "Entity"
    log("generateEntity className:" + className)
    def fields = calcFields(table)
    def packageName = sdkModulePackage + ".dao.entity"

    def path = new File(dir + getPackagePath(packageName))
    if (!path.exists()) {
        path.mkdirs();
    }
    File file = new File(path, className + ".java");
    if (file.exists() && false)
        exceptions.add("Entity file exists. " + file.getName());
    else {
        PrintWriter output = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file), "utf-8"))
        output.withPrintWriter { out -> generateEntity(out, table.getName(), packageName, className, fields) }
    }

    return packageName + "." + className;
}

def generateEntity(out, table, packageName, className, fields) {
    //判断是否包含通用字段
    commonFields = ["status", "note", "sort", "version", "createdBy", "createdAt", "updatedBy", "updatedAt", "deletedFlag", "deletedAt", "deletedBy"];
    def commonFieldCount = 0;
    fields.each() {
        if (commonFields.contains(it.name))
            commonFieldCount++;
    }
    //包含的数据类型，便于引入对应的包
    Set types = new HashSet()
    if (commonFieldCount >= 8)
        fields.each() {
            if (!commonFields.contains(it.name))
                types.add(it.type)
        }
    else
        fields.each() {
            types.add(it.type)
        }

    //start
    out.println """package ${packageName};

import lombok.Data;
import lombok.EqualsAndHashCode;
import org.springframework.data.jpa.domain.support.AuditingEntityListener;
import javax.persistence.*;
import java.io.Serializable;"""

    //是否继承通用字段类
    if (commonFieldCount >= 8)
        out.println "import com.qrqy.developer.common.entity.CommonEntity;"
    else
        out.println "import com.qrqy.developer.common.entity.BaseEntity;"

    //引入类型包
    if (types.contains('LocalDateTime') || types.contains('LocalDate') || types.contains('LocalTime')) {
        out.println 'import java.time.*;'
        out.println 'import com.fasterxml.jackson.databind.annotation.JsonDeserialize;'
        out.println 'import com.fasterxml.jackson.databind.annotation.JsonSerialize;'
        if (types.contains('LocalDate')) {
            out.println 'import com.fasterxml.jackson.datatype.jsr310.deser.LocalDateDeserializer;'
            out.println 'import com.fasterxml.jackson.datatype.jsr310.ser.LocalDateSerializer;'
        } else if (types.contains('LocalTime')) {
            out.println 'import com.fasterxml.jackson.datatype.jsr310.deser.LocalTimeDeserializer;'
            out.println 'import com.fasterxml.jackson.datatype.jsr310.ser.LocalTimeSerializer;'
        } else if (types.contains('LocalDateTime')) {
            out.println 'import com.fasterxml.jackson.datatype.jsr310.deser.LocalDateTimeDeserializer;'
            out.println 'import com.fasterxml.jackson.datatype.jsr310.ser.LocalDateTimeSerializer;'
        }
    }
    if (types.contains('BigDecimal')) {
        out.println 'import java.math.BigDecimal;'
    }
    if (types.contains('InputStream')) {
        out.println 'import java.io.InputStream;'
    }

    out.println """
/**
 * @author : jCoder
 * @date : ${getCurrentTime()}
 */
@Data
@EqualsAndHashCode(callSuper = true)
@Entity
@Table(name = "${table}")
@EntityListeners(AuditingEntityListener.class)
public class ${className} extends ${commonFieldCount >= 8 ? "CommonEntity" : "BaseEntity"} implements Serializable{

    private static final long serialVersionUID = ${getRandom()}L;
"""
    fields.each() {
        if (!commonFields.contains(it.name) || commonFieldCount < 8) {
            // 输出注释
            if (it.comment != null && !it.comment.toString().trim().equals("")) {
                out.println "    /**"
                out.println "     * ${it.comment.toString()}"
                out.println "     */"
            }
            if (it.key) {
                out.println "    @Id";
                out.println "    @GeneratedValue(strategy = GenerationType.IDENTITY)";
            }
            switch (it.type.toString().trim()) {
                case 'LocalDate':
                    out.println "    @JsonDeserialize(using = LocalDateDeserializer.class)"
                    out.println "    @JsonSerialize(using = LocalDateSerializer.class)"
                    break;
                case 'LocalTime':
                    out.println "    @JsonDeserialize(using = LocalTimeDeserializer.class)"
                    out.println "    @JsonSerialize(using = LocalTimeSerializer.class)"
                    break;
                case 'LocalDateTime':
                    out.println "    @JsonDeserialize(using = LocalDateTimeDeserializer.class)"
                    out.println "    @JsonSerialize(using = LocalDateTimeSerializer.class)"
                    break;
            }
            if (it.annos != "") out.println "    ${it.annos}"
            out.println "    private ${it.type} ${it.name};"
            out.println ''
        }
    }


    out.println """
/**
  * JPA版本号，用于增加@Version锁机制
  */
@Column(name = "version")
protected Integer version;
    """

    out.println ""
    out.println "}"
}


/**
 * 生成Vo
 * @param table
 * @param dir
 */
def generateVo(table, dir, voName) {
    def className = camelCase(table.getName(), true) + voName

    def fields = calcFields(table)
    def packageName = sdkModulePackage + ".model.vo"

    def path = new File(dir + getPackagePath(packageName))
    if (!path.exists()) {
        path.mkdirs();
    }
    File file = new File(path, className + ".java");
    if (file.exists() && false)
        exceptions.add("Vo file exists. " + file.getName());
    else {
        PrintWriter output = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file), "utf-8"))
        output.withPrintWriter { out -> generateVo(out, table.getName(), packageName, className, fields) }
    }

    return packageName + "." + className;
}

def generateVo(out, table, packageName, className, fields) {
    //判断是否包含通用字段
    commonFields = ["updatedBy", "updatedAt"];

    //包含的数据类型，便于引入对应的包
    Set types = new HashSet()
    fields.each() {
        log("it.name:" + it.name)
        if (!commonFields.contains(it.name))
            types.add(it.type)
    }

//start
    out.println """package ${packageName};
import io.swagger.annotations.ApiModelProperty;
import lombok.Data;
import java.io.Serializable;"""
//引入类型包
    if (types.contains('LocalDateTime') || types.contains('LocalDate') || types.contains('LocalTime')) {
        out.println 'import java.time.*;'
        out.println 'import com.fasterxml.jackson.databind.annotation.JsonDeserialize;'
        out.println 'import com.fasterxml.jackson.databind.annotation.JsonSerialize;'
    }
    if (types.contains('BigDecimal')) {
        out.println 'import java.math.BigDecimal;'
    }
    if (types.contains('InputStream')) {
        out.println 'import java.io.InputStream;'
    }

    out.println """
/**
 * @author : jCoder
 * @date : ${getCurrentTime()}
 */
@Data
public class ${className} implements Serializable{
    private static final long serialVersionUID = ${getRandom()}L;
"""

    fields.each() {
        if (!commonFields.contains(it.name)) {
            // 输出注释
            if (it.comment != null && !it.comment.toString().trim().equals("")) {
                out.println "    /**"
                out.println "     * ${it.comment.toString()}"
                out.println "     */"

                out.println "    @ApiModelProperty(value = \"${it.comment.toString()}\", example = \" \")"
            }
            out.println "    private ${it.type} ${it.name};"
            out.println ''
        }
    }

    out.println ""
    out.println "}"
}

/**
 * 生成 Repository
 * @param table
 * @param dir
 * @return
 */
def generateRepository(table, dir, entityClassFullName) {
    def className = camelCase(table.getName(), true)
    className += "Repository"

    def packageName = sdkModulePackage + ".dao.repository"

    def path = new File(dir + getPackagePath(packageName))
    if (!path.exists()) {
        path.mkdirs();
    }

    File file = new File(path, className + ".java");
    if (file.exists() && false)
        exceptions.add("Repository file exists. " + file.getName());
    else {
        PrintWriter output = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file), "utf-8"))
        output.withPrintWriter { out -> generateRepository(out, packageName, className, entityClassFullName) }
    }

    return packageName + "." + className;
}

def generateRepository(out, packageName, className, entityClassFullName) {
    def entityClassName = getClassShortName(entityClassFullName);

    out.println """package ${packageName};

import ${entityClassFullName};
import com.qrqy.developer.sql.IBaseSqlRepository;
import org.springframework.stereotype.Component;

/**
 * @author : Luis
 * @date : ${getCurrentTime()}
 */
@Component
public interface ${className} extends IBaseSqlRepository<${entityClassName}, Integer> {

}"""
}


/**
 * 生成 Manager
 * @param table
 * @param dir
 * @return
 */
def generateManager(table, dir, entityClassFullName, repositoryClassFullName) {
    def className = camelCase(table.getName(), true)
    className += "Manager"

    def packageName = sdkModulePackage + ".manager"

    def path = new File(dir + getPackagePath(packageName))
    if (!path.exists()) {
        path.mkdirs();
    }

    File file = new File(path, className + ".java");
    if (file.exists() && false)
        exceptions.add("Manager file exists. " + file.getName());
    else {
        PrintWriter output = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file), "utf-8"))
        output.withPrintWriter { out -> generateManager(out, packageName, className, entityClassFullName, repositoryClassFullName) }
    }

    return packageName + "." + className;
}

def generateManager(out, packageName, className, entityClassFullName, repositoryClassFullName) {
    def entityClassName = getClassShortName(entityClassFullName)
    def repositoryClassName = getClassShortName(repositoryClassFullName)

    out.println """package ${packageName};

import ${entityClassFullName};
import ${repositoryClassFullName};
import com.qrqy.developer.sql.BaseSqlManager;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;

/**
 * @author : Luis
 * @date : ${new SimpleDateFormat("yyyy-MM-dd HH:mm").format(new Date())}
 */
@Component
public class ${className} extends BaseSqlManager<${repositoryClassName}, ${entityClassName}> {
    @Resource
    private ${repositoryClassName} repository;

    @Override
    protected ${repositoryClassName} getRepository() {
        return repository;
    }

    //todo other method

}"""
}

/**
 * 生成类注释
 * @param out
 */
def generateComment(out) {
    out.println "/**"
    out.println " * @author : jCoder"
    out.println " * @date : " + new SimpleDateFormat("yyyy-MM-dd HH:mm").format(new Date())
    out.println " */"
}


def calcFields(table) {
    def idFound = 0;
    DasUtil.getColumns(table).reduce([]) { fields, col ->
        def spec = Case.LOWER.apply(col.getDataType().getSpecification())
        def typeStr = typeMapping.find { p, t -> p.matcher(spec).find() }.value
        def fieldName = camelCase(col.getName(), false);
        fields += [
                name   : fieldName,
                colName: col.getName(),
                type   : typeStr,
                key    : (fieldName.endsWith("Id") || fieldName.endsWith("id")) && idFound++ == 0,
                comment: col.getComment(),
                annos  : "@Column(name = \"" + col.getName() + "\")"]
    }
}

def camelCase(str, capitalize) {
    str = str.replaceAll("ibd_", "").replaceAll("sy_", "").replaceAll("dp_", "");
    def s = com.intellij.psi.codeStyle.NameUtil.splitNameIntoWords(str)
            .collect { Case.LOWER.apply(it).capitalize() }
            .join("")
            .replaceAll(/[^\p{javaJavaIdentifierPart}[_]]/, "_")
    s = capitalize || s.length() == 1 ? s : Case.LOWER.apply(s[0]) + s[1..-1]
    return s;
}


def getClassShortName(classFullName) {
    String[] qoArray = classFullName.split("\\.")
    return qoArray[qoArray.length - 1]
    //return classFullName.replaceAll(".*\\.", "")
}

def getCurrentTime() {
    new SimpleDateFormat("yyyy-MM-dd HH:mm").format(new Date())
}

def getRandom() {
    Math.abs(new Random().nextLong())
}

def getPackagePath(packageName) {
    String[] vars = packageName.split("\\.")
    File.separator + Arrays.stream(vars).collect(Collectors.joining(File.separator));
}

/**
 * 版本管理，暂时无意义
 * @param curVersion 当前版本
 * @return true/false
 */
def checkVersion(curVersion) {
    return true;
}